#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#define BUFSIZE 1024

double ltod(long l) {
  unsigned long long ll = l;
  ll = ll & 0x000000007F800000LL ? ll << 32 & 0x8000000000000000LL
    | ll - 0x0000000000800000LL << 32 & 0x4000000000000000LL
    | ~ll + 0x0000000000800000LL << 31 & 0x2000000000000000LL
    | ~ll + 0x0000000000800000LL << 30 & 0x1000000000000000LL
    | ~ll + 0x0000000000800000LL << 29 & 0x0800000000000000LL
    | ll - 0x0000000000800000LL << 29 & 0x07FFFFFFE0000000LL :
    0x0000000000000000LL;
  return *(double *)&ll;
}

unsigned long dtol(double d) {
  unsigned long long ll = *(unsigned long long *)&d;
  ll = ll >> 32 & 0x0000000080000000LL
    | ll + 0x0010000000000000LL >> 32 & 0x0000000040000000LL
    | ll + 0x0010000000000000LL >> 29 & 0x000000003FFFFFFFLL;
  return (unsigned long)ll;
}

double unorm(double ld, double rd) {
  ld = ld < 0 ? -ld : ld;
  rd = rd < 0 ? -rd : rd;
  return ld < rd ? rd : ld;
}

/* Period parameters */  
#define N 624
#define M 397
/* constant vector a */
#define MATRIX_A 0x9908b0dfUL
/* most significant w-r bits */
#define UPPER_MASK 0x80000000UL
/* least significant r bits */
#define LOWER_MASK 0x7fffffffUL
/* the array for the state vector  */
static unsigned long mt[N]; 
/* mti==N+1 means mt[N] is not initialized */
static int mti = N + 1; 

/* initializes mt[N] with a seed */
void mtsrand(unsigned long s)
{
  mt[0]= s;
  for (mti = 1; mti < N; mti++) {
    mt[mti] = 1812433253UL * (mt[mti - 1] ^ mt[mti - 1] >> 30) + mti;
    /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
    /* In the previous versions, MSBs of the seed affect   */
    /* only MSBs of the array mt[].                        */
    /* 2002/01/09 modified by Makoto Matsumoto             */
  }
}

/* generates a random number on [0,0xffffffff]-interval */
unsigned long mtrand(void)
{
  unsigned long y;
  /* mag01[x] = x * MATRIX_A  for x=0,1 */  
  static unsigned long mag01[2] = {0UL, MATRIX_A};
  /* generate N words at one time */
  if (mti >= N) { 
    int kk;
    /* if mtsrand() has not been called, 
       a default initial seed is used */
    if (mti == N + 1) mtsrand(5489UL);
    for (kk = 0; kk < N - M; kk++) {
      y = mt[kk] & UPPER_MASK | mt[kk + 1] & LOWER_MASK;
      mt[kk] = mt[kk + M] ^ y >> 1 ^ mag01[y & 0x1UL];
    }
    for (; kk < N - 1; kk++) {
      y = mt[kk] & UPPER_MASK | mt[kk + 1] & LOWER_MASK;
      mt[kk] = mt[kk + M - N] ^ y >> 1 ^ mag01[y & 0x1UL];
    }
    y = mt[N - 1] & UPPER_MASK | mt[0] & LOWER_MASK;
    mt[N - 1] = mt[M - 1] ^ y >> 1 ^ mag01[y & 0x1UL];
    mti = 0;
  }
  y = mt[mti++];  
  /* Tempering */
  y ^= y >> 11;
  y ^= y << 7 & 0x9d2c5680UL;
  y ^= y << 15 & 0xefc60000UL;
  y ^= y >> 18;
  return y;
}

int main(int argc, char *argv[]) {
  int i;
  unsigned long term[BUFSIZE << 1];
  unsigned long lterml, rterml;
  unsigned long pmull, tmull;
  unsigned long taddl, paddl;
  unsigned long tsubl, psubl;
  unsigned long tdivl, pdivl;
  unsigned long tsqrtl, psqrtl;
  double ltermd, rtermd;
  double pmuld, tmuld;
  double taddd, paddd;
  double tsubd, psubd;
  double tdivd, pdivd;
  double tsqrtd, psqrtd;
  double tmule, pmule;
  double tadde, padde;
  double tsube, psube;
  double tdive, pdive;
  double tsqrte, psqrte;
  double mule, mule_min = 0, mule_max = 0;
  double adde, adde_min = 0, adde_max = 0;
  double sube, sube_min = 0, sube_max = 0;
  double dive, dive_min = 0, dive_max = 0;
  double sqrte, sqrte_min = 0, sqrte_max = 0;
  int report, abort;
  FILE *fp;
  mtsrand(time(NULL));
  printf("ERROR EVALUATION REPORT\n");
 loop:
  /* prepare */
  for (i = 0; i < BUFSIZE; i++) {
    term[i << 1] = mtrand();
    term[i << 1 | 1] = mtrand();
  }
  for (i = 0; i < BUFSIZE << 1; i++) {
    if (!(term[i] & 0x7F800000L)) term[i] &= 0x80000000L;
  }
  fp = fopen("input.bin", "wb");
  fwrite(term, 4, BUFSIZE << 1, fp);
  fclose(fp);
  
  /* calculate */
  if (system("./simulate float.bin < input.bin > output.bin")) return 1;
  // if (system("./communicate -t 100 < input.bin > output.bin 2> /dev/null")) return 1;
  
  /* evaluate */
  fp = fopen("output.bin", "rb");
  for (i = 0; i < BUFSIZE; i++) {
    lterml = term[i << 1];
    rterml = term[i << 1 | 1];
    ltermd = ltod(lterml);
    rtermd = ltod(rterml);
    fread(&pmull, 4, 1, fp);
    fread(&paddl, 4, 1, fp);
    fread(&psubl, 4, 1, fp);
    fread(&pdivl, 4, 1, fp);
    fread(&psqrtl, 4, 1, fp);
    tmuld = ltermd * rtermd;
    taddd = ltermd + rtermd;
    tsubd = ltermd - rtermd;
    tdivd = ltermd / rtermd;
    tsqrtd = sqrt(ltermd);
    pmuld = ltod(pmull);
    paddd = ltod(paddl);
    psubd = ltod(psubl);
    pdivd = ltod(pdivl);
    psqrtd = ltod(psqrtl);
    tmull = dtol(tmuld);
    taddl = dtol(taddd);
    tsubl = dtol(tsubd);
    tdivl = dtol(tdivd);
    tsqrtl = dtol(tsqrtd);
    pmule = ltermd > -1.70141183460469e+38 &&  ltermd < 1.70141183460469e+38
      && rtermd > -1.70141183460469e+38 &&  rtermd < 1.70141183460469e+38
      && tmuld > -1.70141183460469e+38 &&  tmuld < 1.70141183460469e+38 ?
      tmuld < 0 ? tmuld - pmuld : pmuld - tmuld : 0;
    padde = ltermd > -1.70141183460469e+38 &&  ltermd < 1.70141183460469e+38
      && rtermd > -1.70141183460469e+38 &&  rtermd < 1.70141183460469e+38
      && taddd > -1.70141183460469e+38 &&  taddd < 1.70141183460469e+38 ?
      taddd < 0 ? taddd - paddd : paddd - taddd : 0;
    psube = ltermd > -1.70141183460469e+38 &&  ltermd < 1.70141183460469e+38
      && rtermd > -1.70141183460469e+38 &&  rtermd < 1.70141183460469e+38
      && tsubd > -1.70141183460469e+38 &&  tsubd < 1.70141183460469e+38 ?
      tsubd < 0 ? tsubd - psubd : psubd - tsubd : 0;
    pdive = ltermd > -1.70141183460469e+38 &&  ltermd < 1.70141183460469e+38
      && rtermd > -1.70141183460469e+38 &&  rtermd < 1.70141183460469e+38
      && tdivd > -1.70141183460469e+38 &&  tdivd < 1.70141183460469e+38 ?
      tdivd < 0 ? tdivd - pdivd : pdivd - tdivd : 0;
    psqrte = ltermd >= 0 &&  ltermd < 1.70141183460469e+38 ? 
      tsqrtd < 0 ? tsqrtd - psqrtd : psqrtd - tsqrtd : 0;
    tmule = unorm(tmuld * 2.38418579101562e-07, 1.17549435082229e-38);
    tadde = unorm(unorm(unorm(ltermd * 1.19209289550781e-07, rtermd * 1.19209289550781e-07), taddd * 1.19209289550781e-07), 1.17549435082229e-38);
    tsube = unorm(unorm(unorm(ltermd * 1.19209289550781e-07, rtermd * 1.19209289550781e-07), tsubd * 1.19209289550781e-07), 1.17549435082229e-38);
    tdive = unorm(tdivd * 9.5367431640625e-07, 1.17549435082229e-38);
    tsqrte = unorm(tsqrtd * 9.5367431640625e-07, 1.17549435082229e-38);
    mule = pmule * 100. / tmule;
    adde = padde * 100. / tadde;
    sube = psube * 100. / tsube;
    dive = tdive == tdive ? pdive * 100. / tdive : 0;
    sqrte = tsqrte == tsqrte ? psqrte * 100. / tsqrte : 0;
    report = 0;
    abort = 0;
    if (mule > mule_max) {
      mule_max = mule;
      report = 1;
      abort |= mule >= 100.;
    } else if (mule < mule_min) {
      mule_min = mule;
      report |= 1;
      abort |= mule <= -100.;
    }
    if (adde > adde_max) {
      adde_max = adde;
      report |= 2;
      abort |= adde >= 100.;
    } else if (adde < adde_min) {
      adde_min = adde;
      report |= 2;
      abort |= adde <= -100.;
    }
    if (sube > sube_max) {
      sube_max = sube;
      report |= 4;
      abort |= sube >= 100.;
    } else if (sube < sube_min) {
      sube_min = sube;
      report |= 4;
      abort |= sube <= -100.;
    }
    if (dive > dive_max) {
      dive_max = dive;
      report |= 8;
      abort |= dive >= 100.;
    } else if (dive < dive_min) {
      dive_min = dive;
      report |= 8;
      abort |= dive <= -100.;
    }
    if (sqrte > sqrte_max) {
      sqrte_max = sqrte;
      report |= 16;
      abort |= sqrte >= 100.; 
    } else if (sqrte < sqrte_min) {
      sqrte_min = sqrte;
      report |= 16;
      abort |= sqrte <= -100.;
    }
    if (report) {
      printf("========================================================================\n");
      printf("term  : %lx %02lx %06lx (%13e) , %lx %02lx %06lx (%13e)\n",
	     lterml >> 31 & 0x00000001L, lterml >> 23 & 0x000000FFL,
	     lterml & 0x007FFFFFL, ltermd,
	     rterml >> 31 & 0x00000001L, rterml >> 23 & 0x000000FFL,
	     rterml & 0x007FFFFFL, rtermd);
    }
    if (report & 1) {
      printf("mul   : %lx %02lx %06lx (%13e) , %lx %02lx %06lx (%13e)\n",
	     pmull >> 31 & 0x00000001L, pmull >> 23 & 0x000000FFL,
	     pmull & 0x007FFFFFL, pmuld,
	     tmull >> 31 & 0x00000001L, tmull >> 23 & 0x000000FFL,
	     tmull & 0x007FFFFFL, tmuld);
      printf("                %13le / %13le (%lf%%) \n", pmule, tmule, mule);
    }
    if (report & 2) {
      printf("add   : %lx %02lx %06lx (%13e) , %lx %02lx %06lx (%13e)\n",
	     paddl >> 31 & 0x00000001L, paddl >> 23 & 0x000000FFL,
	     paddl & 0x007FFFFFL, paddd,
	     taddl >> 31 & 0x00000001L, taddl >> 23 & 0x000000FFL,
	     taddl & 0x007FFFFFL, taddd);
      printf("                %13le / %13le (%lf%%) \n", padde, tadde, adde);
    }
    if (report & 4) {
      printf("sub   : %lx %02lx %06lx (%13e) , %lx %02lx %06lx (%13e)\n",
	     psubl >> 31 & 0x00000001L, psubl >> 23 & 0x000000FFL,
	     psubl & 0x007FFFFFL, psubd,
	     tsubl >> 31 & 0x00000001L, tsubl >> 23 & 0x000000FFL,
	     tsubl & 0x007FFFFFL, tsubd);
      printf("                %13le / %13le (%lf%%) \n", psube, tsube, sube);
    }
    if (report & 8) {
      printf("div   : %lx %02lx %06lx (%13e) , %lx %02lx %06lx (%13e)\n",
	     pdivl >> 31 & 0x00000001L, pdivl >> 23 & 0x000000FFL,
	     pdivl & 0x007FFFFFL, pdivd,
	     tdivl >> 31 & 0x00000001L, tdivl >> 23 & 0x000000FFL,
	     tdivl & 0x007FFFFFL, tdivd);
      printf("                %13le / %13le (%lf%%) \n", pdive, tdive, dive);
    }
    if (report & 16) {
      printf("sqrt  : %lx %02lx %06lx (%13e) , %lx %02lx %06lx (%13e)\n",
	     psqrtl >> 31 & 0x00000001L, psqrtl >> 23 & 0x000000FFL,
	     psqrtl & 0x007FFFFFL, psqrtd,
	     tsqrtl >> 31 & 0x00000001L, tsqrtl >> 23 & 0x000000FFL,
	     tsqrtl & 0x007FFFFFL, tsqrtd);
      printf("                %13le / %13le (%lf%%) \n", psqrte, tsqrte, sqrte);
    }
    if (abort) return 0;
  }
  fclose(fp);
  goto loop;
}

