#include <stdio.h>
#include <math.h>

#define OP_ALU 0x0UL

#define FNCT_ADD 0x0UL
#define FNCT_SUB 0x1UL
#define FNCT_SLL 0x2UL
#define FNCT_SRR 0x3UL
#define FNCT_AND 0x4UL
#define FNCT_OR  0x5UL
#define FNCT_XOR 0x6UL
#define FNCT_NOT 0x7UL
#define FNCT_FADD  0x8UL
#define FNCT_FSUB  0x9UL
#define FNCT_FMUL  0xAUL
#define FNCT_FDIV  0xCUL
#define FNCT_FSQRT 0xDUL
#define FNCT_RD 0xEUL
#define FNCT_WR 0xFUL

#define FNCT_SIN 0x10UL
#define FNCT_COS 0x11UL
#define FNCT_ATAN 0x12UL

#define OP_ADDI 0x1UL

#define OP_LOAD  0x2UL
#define OP_STORE 0x3UL

#define OP_JR   0xEUL
#define OP_JAL  0xFUL

#define OP_BL   0xAUL
#define OP_BE   0x8UL
#define OP_BLE  0xBUL
#define OP_BNE  0x9UL
#define OP_FBL   0xCUL
#define OP_FBLE  0xDUL

unsigned long pc;
unsigned long reg[32];
#define MEM_SIZE (1 << 20)
unsigned long mem[MEM_SIZE];

unsigned short S(unsigned long l) {
  return l >> 31;
}
unsigned short E(unsigned long l) {
  return l >> 23 & 0xFF;
}
unsigned long F(unsigned long l) {
  return l & 0x7FFFFF | (l >> 23 & 0xFFL ? 0x800000L : 0L);
}
unsigned long L(unsigned long s, unsigned long e, unsigned long f) {
  int i;
  for (i = 0; i < 25; i++) {
    if (f & 0x1000000UL) break;
    e -= 1UL;
    f <<= 1;
  }
  if (i == 25) e = -1;
  e = e < 0xFFUL ? e + 1UL & 0xFFUL : 0;
  f = f >> 1 & 0x7FFFFFUL;
  return s << 31 | e << 23 | f;
}

int calc()
{  
  const unsigned long op = mem[pc] >> 26;
  const unsigned long rs = mem[pc] >> 21 & 0x1FUL;
  const unsigned long rt = mem[pc] >> 16 & 0x1FUL;
  const unsigned long rd = mem[pc] >> 11 & 0x1FUL;
  const unsigned long shft = mem[pc] >> 6 & 0x1FUL;
  const unsigned long fnct = mem[pc] & 0x3FUL;
  const unsigned long imm = mem[pc] & 0xFFFFUL ^ (mem[pc] & 0x8000L ?  0xFFFF0000UL : 0UL);
  const unsigned long tgt = mem[pc] & 0x1FFFFFL ^ (mem[pc] & 0x100000UL ?  0xFFE00000UL : 0UL);

  switch(op)
    {
      /**********************************/
      /* GENERAL OPERATIONS             */
      /**********************************/
    case OP_ALU:
      switch(fnct)
	{
	case FNCT_ADD:
	  reg[rd] = reg[rs] + reg[rt];
	  return 1;
	case FNCT_SUB:
	  reg[rd] = reg[rs] - reg[rt];
	  return 1;
#ifdef FNCT_MUL
	case FNCT_MUL:
	  reg[rd] = reg[rs] * reg[rt];
	  return 1;
#endif
#ifdef FNCT_DIV
	case FNCT_DIV:
	  reg[rd] = reg[rs] / reg[rt];
	  return 1;
#endif
	case FNCT_SLL:
	  reg[rd] = reg[rs] << shft;
	  return 1;
#ifdef FNCT_SRL
	case FNCT_SRL:
	  reg[rd] = reg[rs] >> shft;
	  return 1;
#endif
	case FNCT_SRR:
	  reg[rd] = (signed)reg[rs] >> shft;
	  return 1;
	case FNCT_AND:
	  reg[rd] = reg[rs] & reg[rt];
	  return 1;
	case FNCT_OR:
	  reg[rd] = reg[rs] | reg[rt];
	  return 1;
	case FNCT_XOR:
	  reg[rd] = reg[rs] ^ reg[rt];
	  return 1;
	case FNCT_NOT:
	  reg[rd] = ~reg[rs];
	  return 1;
	case FNCT_FADD:
	  if (E(reg[rs]) > E(reg[rt])
	      || E(reg[rs]) == E(reg[rt]) && F(reg[rs]) >= F(reg[rt])) {
	    reg[rd]
	      = L(S(reg[rs]),
		  E(reg[rs]),
		  E(reg[rs]) - E(reg[rt]) > 23 ?
		  F(reg[rs]) :
		  S(reg[rs]) ^ S(reg[rt]) ?
		  F(reg[rs]) - (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])) :
		  F(reg[rs]) + (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])));
	  } else {
	    reg[rd]
	      = L(S(reg[rt]),
		  E(reg[rt]), 
		  E(reg[rt]) - E(reg[rs]) > 23 ?
		  F(reg[rt]) :
		  S(reg[rs]) ^ S(reg[rt]) ?
		  F(reg[rt]) - (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])) :
		  F(reg[rt]) + (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])));
	  }
	  return 1;
	case FNCT_FSUB:
	  if (E(reg[rs]) > E(reg[rt])
	      || E(reg[rs]) == E(reg[rt]) && F(reg[rs]) >= F(reg[rt])) {
	    reg[rd]
	      = L(S(reg[rs]),
		  E(reg[rs]),
		  E(reg[rs]) - E(reg[rt]) > 23 ?
		  F(reg[rs]) :
		  S(reg[rs]) ^ S(reg[rt]) ?
		  F(reg[rs]) + (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])) :
		  F(reg[rs]) - (F(reg[rt]) >> E(reg[rs]) - E(reg[rt])));
	  } else {
	    reg[rd]
	      = L(!S(reg[rt]),
		  E(reg[rt]),
		  E(reg[rt]) - E(reg[rs]) > 23 ?
		  F(reg[rt]) :
		  S(reg[rs]) ^ S(reg[rt]) ?
		  F(reg[rt]) + (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])) :
		  F(reg[rt]) - (F(reg[rs]) >> E(reg[rt]) - E(reg[rs])));
	  }
	  return 1;
	case FNCT_FMUL:
	  reg[rd]
	    = L(S(reg[rs]) ^ S(reg[rt]),
		E(reg[rs]) + E(reg[rt]) - 128,
		(((long long)F(reg[rt]) * (F(reg[rs]) % 0x1000UL)) >> 23)
		+ (((long long)F(reg[rt]) * (F(reg[rs]) / 0x1000UL)) >> 11));
	  return 1;
	case FNCT_FDIV:
	  {
	    const unsigned long addr
	      = F(reg[rs]) >> 13 & 0x3FEUL | 0x401UL;
	    const unsigned long long tbl
	      = 0x400000ULL / addr - 0x200000ULL / addr & 0x3FFUL | 0x400UL;
	    reg[rd]
	      = L(S(reg[rs]),
		  255 - E(reg[rs]),
		  tbl * (0x1000006UL - F(reg[rs]) / 2 * tbl / 0x400UL) / 0x400UL);	    
	  }
	  return 1;
	case FNCT_FSQRT:
	  if (F(reg[rs]) & 0x800000) {
	    const unsigned long addr
	      = (F(reg[rs]) >> 14 & 0x1FEUL | 0x201UL) << (E(reg[rs]) & 0x1UL);
	    const unsigned long long itbl
	      = (int)(1. / sqrt((double)addr / 0x200UL) * 0x1000UL + 0.5) >> 1 & 0x3FF | 0x400;
	    const unsigned long long tbl
	      = 0x100000000ULL / itbl - 0x80000000ULL / itbl - 1 & 0x0FFFFFUL | 0x100000;
	    reg[rd]
	      = L(0,
		  E(reg[rs]) / 2 + 64,
		  tbl * 0x4UL
		  + (F(reg[rs]) / 2 * itbl / 0x800 << E(reg[rs]) % 2));
	  } else {
	    reg[rd] = 0L;
	  }
	  return 1;
	case FNCT_RD:
	  return (reg[rd] = getchar()) != (unsigned)EOF;
	case FNCT_WR:
	  putchar(reg[rs]);
	  return 1;
	case FNCT_SIN:
	  {
	    float f = *(float *)&reg[rs];
	    f = sin(f / 2.) * 2.;
	    reg[rd] = *(long *)&f;
	    return 1;
	  }
	case FNCT_COS:
	  {
	    float f = *(float *)&reg[rs];
	    f = cos(f / 2.) * 2.;
	    reg[rd] = *(long *)&f;
	    return 1;
	  }
	case FNCT_ATAN:
	  {
	    float f = *(float *)&reg[rs];
	    f = atan(f / 2.) * 2.;
	    reg[rd] = *(long *)&f;
	    return 1;
	  }
	default:
	  fprintf(stderr, "Undefined function. (%ld, 0x%08lx)\n", pc, mem[pc]);
	  return 0;
	}
      return 1;
      
    case OP_ADDI:
      reg[rt] = reg[rs] + imm;
      return 1;

      /**********************************/
      /* MEMORY OPERATIONS              */
      /**********************************/
    case OP_LOAD:
      {
	const unsigned long addr = (unsigned)(reg[rs] + imm) % MEM_SIZE;
	if (addr < 256
	    || addr >= 0x40000 && addr < reg[1] - 4
	    || addr >= reg[1] + 256 && addr < 0x8000
	    || addr >= reg[2]) {
	  fprintf(stderr, "Exception. (%ld, 0x%05lx)\n", pc, addr);
	  return 0;
	}
	reg[rt] = mem[addr];
	return 1;
      }
    case OP_STORE:
      {
	const unsigned long addr = (unsigned)(reg[rs] + imm) % MEM_SIZE;
	if (addr < 256
	    || addr >= 0x40000 && addr < reg[1] - 4
	    || addr >= reg[1] + 256 && addr < 0x8000
	    || addr >= reg[2]) {
	  fprintf(stderr, "Exception. (%ld, 0x%05lx)\n", pc, addr);
	  return 0;
	}
	mem[addr] = reg[rt];
	return 1;
      }
      
      /**********************************/
      /* JUMP OPERATIONS                */
      /**********************************/
    case OP_JR:
      //if (rs) printf("ret %d\n", reg[rs] + tgt);
      //else printf("jump %d\n", tgt);

      return reg[rs] + tgt - pc;

    case OP_JAL:
      //printf("call %d\n", pc + tgt);
      reg[31] = pc;		//1 or 4?
      return tgt;
      
      /**********************************/
      /* INT BRANCH OPERATIONS          */
      /**********************************/
    case OP_BL  :
      return (signed)reg[rs] <  (signed)reg[rt] ? imm : 1;
    case OP_BE  :                  
      return (signed)reg[rs] == (signed)reg[rt] ? imm : 1;
    case OP_BLE :                  
      return (signed)reg[rs] <= (signed)reg[rt] ? imm : 1;
    case OP_BNE :                  
      return (signed)reg[rs] != (signed)reg[rt] ? imm : 1;
      /**********************************/
      /* FLOAT BRANCH OPERATIONS        */
      /**********************************/
    case OP_FBL  :
      return (signed)reg[rs] < (signed)reg[rt] ? imm : 1;
    case OP_FBLE :                 
      return (signed)reg[rs] <= (signed)reg[rt] ? imm : 1;

    default:
      fprintf(stderr, "Undefined operation. (%ld, 0x%08lx)\n", pc, mem[pc]);
      return 0;

    }   
  
}

int main(int argc, char *argv[])
{
  FILE *fp;
  int dpc;  

  if(argc != 2 || (fp = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "usage: simulate file\n");
    return 1;
  }
  
  mem[fread(mem, 1, MEM_SIZE, fp)] = OP_BE << 26;
  
  pc = 0;
  reg[0] = 0;
  
  while (dpc = calc()) {
    pc = (pc + dpc) % MEM_SIZE;
  }
  
  return 0;
}


